""" SQLite Backend class built on top of base Backend """

import os
from datetime import datetime
from copy import deepcopy
import threading

import sqlalchemy

from dagobah.backend.base import BaseBackend
from dagobah.backend.sqlite_models import (Base, Dagobah, DagobahJob,
                                           DagobahTask, DagobahDependency,
                                           DagobahLog, DagobahLogTask)


class SQLiteBackend(BaseBackend):
    """ SQLite Backend implementation. """

    def __init__(self, filepath):
        super(SQLiteBackend, self).__init__()

        self.filepath = filepath
        if self.filepath == 'default':
            location = os.path.realpath(os.path.join(os.getcwd(),
                                                     os.path.dirname(__file__)))
            self.filepath = os.path.join(location, 'dagobah.db')

        connect_args = {'check_same_thread': False}
        self.engine = sqlalchemy.create_engine('sqlite:///' + self.filepath,
                                               connect_args=connect_args)
        self.Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = self.Session()

        Base.metadata.create_all(self.engine)

        self.lock = threading.Lock()


    def __repr__(self):
        return '<SQLiteBackend (path: %s)>' % (self.filepath)


    def get_known_dagobah_ids(self):
        results = []
        for rec in self.session.query(Dagobah).all():
            results.append(rec.id)
        return results


    def get_new_dagobah_id(self):
        count = self.session.query(sqlalchemy.func.max(Dagobah.id)).scalar()
        return max(count, 0) + 1


    def get_new_job_id(self):
        count = self.session.query(sqlalchemy.func.max(DagobahJob.id)).scalar()
        return max(count, 0) + 1


    def get_new_log_id(self):
        count = self.session.query(sqlalchemy.func.max(DagobahLog.id)).scalar()
        return max(count, 0) + 1


    def get_dagobah_json(self, dagobah_id):
        return self.session.query(Dagobah).\
            filter_by(id=dagobah_id).\
            one().json


    def commit_dagobah(self, dagobah_json):

        rec = self.session.query(Dagobah).\
            filter_by(id=dagobah_json['dagobah_id']).\
            first()

        if not rec:
            rec = Dagobah()
            self.session.add(rec)

        for job in dagobah_json.get('jobs', []):

            existing = self.session.query(DagobahJob).\
                filter_by(id=job['job_id']).\
                first()

            if existing:
                self._update_job_rec(existing, dagobah_json, 'dagobah')
            else:
                new_job = DagobahJob(job['name'])
                rec.jobs.append(new_job)
                self._update_job_rec(new_job, dagobah_json, 'dagobah')

        self.session.commit()


    def delete_dagobah(self, dagobah_id):
        """ Deletes the Dagobah and all child Jobs from the database.

        Related run logs are deleted as well.
        """

        rec = self.session.query(Dagobah).filter_by(id=dagobah_id).first()
        if not rec:
            raise KeyError('no dagobah doc found with id %s' % dagobah_id)

        for job in rec.jobs:
            self.delete_job(job.id)

        self.session.delete(rec)
        self.session.commit()


    def commit_job(self, job_json):

        rec = self.session.query(DagobahJob).\
            filter_by(id=job_json['job_id']).\
            first()

        if not rec:
            rec = DagobahJob(job_json['name'])
            self.session.add(rec)

        self._update_job_rec(rec, job_json, 'job')
        self.session.commit()


    def delete_job(self, job_id):

        # TODO: get cascading deletes to work automatically with sqlite
        # this should be a generalizable solution for all supported DBs, though

        job = self.session.query(DagobahJob).filter_by(id=job_id).one()

        for task in job.tasks:
            self.session.delete(task)
        for dep in job.dependencies:
            self.session.delete(dep)
        for log in job.logs:
            for task in log.tasks:
                self.session.delete(task)
            self.session.delete(log)
        self.session.delete(job)

        self.session.commit()


    def commit_log(self, log_json):

        rec = self.session.query(DagobahLog).\
            filter_by(id=log_json['log_id']).\
            first()

        if rec:
            rec.update_from_dict(log_json)
        else:
            rec = DagobahLog()
            self.session.add(rec)

        # delete any deprecated logtask records for this log
        for log_task in self.session.query(DagobahLogTask).\
            filter_by(log_id=rec.id).\
            all():
            if log_task.name not in log_json.get('tasks', {}).keys():
                self.session.delete(log_task)

        self.session.flush()

        for task_name, task_data in log_json.get('tasks', {}).iteritems():

            existing = self.session.query(DagobahLogTask).\
                filter_by(log_id=rec.id).\
                filter_by(name=task_name).\
                first()

            if not existing:
                existing = DagobahLogTask(task_name)
                self.session.add(existing)
                rec.tasks.append(existing)

            existing.update_from_dict(task_data)


    def get_latest_run_log(self, job_id, task_name):
        log = self.session.query(DagobahLog).\
            filter_by(job_id=job_id).\
            order_by(DagobahLog.save_date.desc()).\
            first()
        return log.json


    def acquire_lock(self):
        self.lock.acquire()


    def release_lock(self):
        self.lock.release()


    def _update_job_rec(self, job_rec, in_data, data_type):
        """" Update the passed DagobahJob record and its children Tasks.

        Does not commit the session after its updates.
        """

        if data_type not in ['dagobah', 'job']:
            raise KeyError('unknown data_type %s' % data_type)

        data = deepcopy(in_data)

        if data_type == 'dagobah':

            found_job = False
            for job in data.get('jobs', []):
                if job['job_id'] == job_rec.id:
                    data = job
                    found_job = True
                    break

            if not found_job:
                raise KeyError('job %s not found in in_data' % job_rec.name)

        # update the job record itself
        job_rec.update_from_dict(data)

        job_tasks = data.get('tasks', {})
        job_deps = data.get('dependencies', {})

        # clean up tasks and dependencies that have been removed
        for existing_task in self.session.query(DagobahTask).\
                filter_by(job_id=job_rec.id).\
                all():
            if existing_task.name not in job_deps:
                self.session.query(DagobahDependency).\
                    filter(sqlalchemy.or_(DagobahDependency.from_task_id==existing_task.id,
                                          DagobahDependency.to_task_id==existing_task.id)).\
                    delete()
                self.session.delete(existing_task)

        for existing_dep in self.session.query(DagobahDependency).\
                filter_by(job_id=job_rec.id).\
                all():
            if (existing_dep.from_task.name not in job_deps or
                (existing_dep.to_task.name not in
                 job_deps[existing_dep.from_task.name])):
                self.session.delete(existing_dep)

        # update and create tasks
        for task_data in job_tasks:

            existing = self.session.query(DagobahTask).\
                filter_by(job_id=job_rec.id).\
                filter_by(name=task_data['name']).\
                first()

            if not existing:
                existing = DagobahTask(task_data['name'], task_data['command'])
                self.session.add(existing)
                job_rec.tasks.append(existing)

            self._update_task_rec(existing, data)

        # get up-to-date task IDs and create a name-to-ID lookup
        self.session.flush()
        task_lu = {}
        for task in job_rec.tasks:
            task_lu[task.name] = task.id

        # update and create dependencies
        for from_task_name, to_tasks in job_deps.iteritems():
            for to_task_name in to_tasks:
                from_task_id = task_lu[from_task_name]
                to_task_id = task_lu[to_task_name]
                existing = self.session.query(DagobahDependency).\
                    filter_by(job_id=job_rec.id).\
                    filter_by(from_task_id=from_task_id).\
                    filter_by(to_task_id=to_task_id).\
                    first()

                if not existing:
                    existing = DagobahDependency(from_task_id, to_task_id)
                    self.session.add(existing)
                    job_rec.dependencies.append(existing)


    def _update_task_rec(self, task_rec, job_data):
        """ Update the passed DagobahTask from a job data dict. """
        for task in job_data.get('tasks', []):
            if task.get('name', None) == task_rec.name:
                task_rec.update_from_dict(task)



""" HTTP API methods for Dagobah daemon. """

from flask import request, abort
from flask_login import login_required

from dagobah.daemon.daemon import app
from dagobah.daemon.util import validate_dict, api_call

dagobah = app.config['dagobah']

@app.route('/api/jobs', methods=['GET'])
@login_required
@api_call
def get_jobs():
    return dagobah._serialize().get('jobs', {})


@app.route('/api/job', methods=['GET'])
@login_required
@api_call
def get_job():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    return job._serialize()


@app.route('/api/head', methods=['GET'])
@login_required
@api_call
def head_task():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str,
                         stream=str,
                         num_lines=int):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    call_args = {}
    for key in ['stream', 'num_lines']:
        if key in args:
            call_args[key] = args[key]
    return task.head(**call_args)


@app.route('/api/tail', methods=['GET'])
@login_required
@api_call
def tail_task():
    args = dict(request.args)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str,
                         stream=str,
                         num_lines=int):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    call_args = {}
    for key in ['stream', 'num_lines']:
        if key in args:
            call_args[key] = args[key]
    return task.tail(**call_args)


@app.route('/api/add_job', methods=['POST'])
@login_required
@api_call
def add_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    dagobah.add_job(args['job_name'])


@app.route('/api/delete_job', methods=['POST'])
@login_required
@api_call
def delete_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    dagobah.delete_job(args['job_name'])


@app.route('/api/start_job', methods=['POST'])
@login_required
@api_call
def start_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.start()


@app.route('/api/retry_job', methods=['POST'])
@login_required
@api_call
def retry_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.retry()


@app.route('/api/add_task_to_job', methods=['POST'])
@login_required
@api_call
def add_task_to_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_command', 'task_name'],
                         job_name=str,
                         task_command=str,
                         task_name=str):
        abort(400)

    dagobah.add_task_to_job(args['job_name'],
                            args['task_command'],
                            args['task_name'])


@app.route('/api/delete_task', methods=['POST'])
@login_required
@api_call
def delete_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.delete_task(args['task_name'])


@app.route('/api/add_dependency', methods=['POST'])
@login_required
@api_call
def add_dependency():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name',
                                   'from_task_name',
                                   'to_task_name'],
                         job_name=str,
                         from_task_name=str,
                         to_task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.add_dependency(args['from_task_name'], args['to_task_name'])


@app.route('/api/delete_dependency', methods=['POST'])
@login_required
@api_call
def delete_dependency():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name',
                                   'from_task_name',
                                   'to_task_name'],
                         job_name=str,
                         from_task_name=str,
                         to_task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.delete_dependency(args['from_task_name'], args['to_task_name'])


@app.route('/api/schedule_job', methods=['POST'])
@login_required
@api_call
def schedule_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'cron_schedule'],
                         job_name=str,
                         cron_schedule=str):
        abort(400)

    if args['cron_schedule'] == '':
        args['cron_schedule'] = None

    job = dagobah.get_job(args['job_name'])
    job.schedule(args['cron_schedule'])


@app.route('/api/stop_scheduler', methods=['POST'])
@login_required
@api_call
def stop_scheduler():
    dagobah.scheduler.stop()


@app.route('/api/restart_scheduler', methods=['POST'])
@login_required
@api_call
def restart_scheduler():
    dagobah.scheduler.restart()


@app.route('/api/terminate_all_tasks', methods=['POST'])
@login_required
@api_call
def terminate_all_tasks():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.terminate_all()


@app.route('/api/kill_all_tasks', methods=['POST'])
@login_required
@api_call
def kill_all_tasks():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    job.kill_all()


@app.route('/api/terminate_task', methods=['POST'])
@login_required
@api_call
def terminate_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)
    task.terminate()


@app.route('/api/kill_task', methods=['POST'])
@login_required
@api_call
def kill_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)
    task.kill()


@app.route('/api/edit_job', methods=['POST'])
@login_required
@api_call
def edit_job():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name'],
                         job_name=str,
                         name=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    del args['job_name']
    job.edit(**args)


@app.route('/api/edit_task', methods=['POST'])
@login_required
@api_call
def edit_task():
    args = dict(request.form)
    if not validate_dict(args,
                         required=['job_name', 'task_name'],
                         job_name=str,
                         task_name=str,
                         name=str,
                         command=str):
        abort(400)

    job = dagobah.get_job(args['job_name'])
    task = job.tasks.get(args['task_name'], None)
    if not task:
        abort(400)

    del args['job_name']
    del args['task_name']
    job.edit_task(task.name, **args)



""" Views for Dagobah daemon. """

from flask import render_template, redirect, url_for
from flask_login import login_required

from dagobah.daemon.daemon import app
from dagobah.daemon.api import get_jobs

dagobah = app.config['dagobah']


@app.route('/', methods=['GET'])
def index_route():
    """ Redirect to the dashboard. """
    return redirect(url_for('jobs'))


@app.route('/jobs', methods=['GET'])
@login_required
def jobs():
    """ Show information on all known Jobs. """
    return render_template('jobs.html',
                           jobs=get_jobs())


@app.route('/job/<job_id>', methods=['GET'])
@login_required
def job_detail(job_id=None):
    """ Show a detailed description of a Job's status. """
    jobs = get_jobs()
    job = [job for job in jobs if str(job['job_id']) == job_id][0]
    return render_template('job_detail.html', job=job)


@app.route('/job/<job_id>/<task_name>', methods=['GET'])
@login_required
def task_detail(job_id=None, task_name=None):
    """ Show a detailed description of a specific task. """
    jobs = get_jobs()
    job = [job for job in jobs if str(job['job_id']) == job_id][0]
    return render_template('task_detail.html', job=job, task_name=task_name)



""" SQLite model definitions. """

from datetime import datetime
from collections import defaultdict

from sqlalchemy import Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

STREAM_LOG_SIZE = 1000000

class Dagobah(Base):
    __tablename__ = 'dagobah'

    id = Column(Integer, primary_key=True)
    created_jobs = Column(Integer, nullable=False)

    jobs = relationship('DagobahJob', backref='parent')

    def __init__(self):
        self.created_jobs = 0


    def __repr__(self):
        return "<SQLite:Dagobah (%d)>" % self.id


    @property
    def json(self):
        return {'dagobah_id': self.id,
                'created_jobs': self.created_jobs,
                'jobs': [job.json for job in self.jobs]}


class DagobahJob(Base):
    __tablename__ = 'dagobah_job'

    id = Column(Integer, primary_key=True)
    parent_id = Column(Integer, ForeignKey('dagobah.id'), index=True)
    name = Column(String(1000))
    status = Column(String(30), nullable=False)
    cron_schedule = Column(String(100))
    next_run = Column(DateTime)

    tasks = relationship('DagobahTask', backref='job')
    dependencies = relationship('DagobahDependency', backref='job')
    logs = relationship('DagobahLog', backref='job')

    def __init__(self, name):
        self.name = name
        self.status = 'waiting'


    def __repr__(self):
        return "<SQLite:DagobahJob (%d)>" % self.id


    @property
    def json(self):
        return {'job_id': self.id,
                'name': self.name,
                'parent_id': self.parent.id,
                'status': self.status,
                'cron_schedule': self.cron_schedule,
                'next_run': self.next_run,
                'tasks': [task.json for task in self.tasks],
                'dependencies': self._gather_dependencies()}


    def update_from_dict(self, data):
        for key in ['parent_id', 'name', 'status', 'cron_schedule',
                    'next_run']:
            if key in data:
                setattr(self, key, data[key])


    def _gather_dependencies(self):
        result = defaultdict(list)
        for dep in self.dependencies:
            result[dep.from_task.name].append(dep.to_task.name)
        return result


class DagobahTask(Base):
    __tablename__ = 'dagobah_task'

    id = Column(Integer, primary_key=True)
    job_id = Column(Integer, ForeignKey('dagobah_job.id'), index=True)
    name = Column(String(1000), nullable=False)
    command = Column(String(1000), nullable=False)
    started_at = Column(DateTime)
    completed_at = Column(DateTime)
    success = Column(String(30))

    def __init__(self, name, command):
        self.name = name
        self.command = command


    def __repr__(self):
        return "<SQLite:DagobahTask (%d)>" % self.id


    @property
    def json(self):
        return {'name': self.name,
                'command': self.command,
                'started_at': self.started_at,
                'completed_at': self.completed_at,
                'success': self.success}


    def update_from_dict(self, data):
        for key in ['job_id', 'name', 'command', 'started_at',
                    'completed_at', 'success']:
            if key in data:
                setattr(self, key, data[key])


class DagobahDependency(Base):
    __tablename__ = 'dagobah_dependency'

    id = Column(Integer, primary_key=True)
    job_id = Column(Integer, ForeignKey('dagobah_job.id'), index=True)
    from_task_id = Column(Integer, ForeignKey('dagobah_task.id'), index=True)
    to_task_id = Column(Integer, ForeignKey('dagobah_task.id'), index=True)

    from_task = relationship('DagobahTask', foreign_keys="DagobahDependency.from_task_id")
    to_task = relationship('DagobahTask', foreign_keys="DagobahDependency.to_task_id")

    def __init__(self, from_task_id, to_task_id):
        self.from_task_id = from_task_id
        self.to_task_id = to_task_id


    def __repr__(self):
        return "<SQLite:DagobahDependency (%d)>" % self.id


    def update_from_dict(self, data):
        for key in ['job_id']:
            if key in data:
                setattr(self, key, data[key])


class DagobahLog(Base):
    __tablename__ = 'dagobah_log'

    id = Column(Integer, primary_key=True)
    job_id = Column(Integer, ForeignKey('dagobah_job.id'), index=True)
    start_time = Column(DateTime)
    last_retry_time = Column(DateTime)
    save_date = Column(DateTime)

    tasks = relationship('DagobahLogTask', backref='log')

    def __init__(self):
        self.save_date = datetime.utcnow()


    def __repr__(self):
        return "<SQLite:DagobahLog (%d)>" % self.id


    @property
    def json(self):
        return {'log_id': self.id,
                'job_id': self.job_id,
                'start_time': self.start_time,
                'name': self.job.name,
                'parent_id': self.job.parent_id,
                'tasks': {task.name: task.json
                          for task in self.tasks}}


    def update_from_dict(self, data):
        for key in ['job_id', 'start_time', 'last_retry_time']:
            if key in data:
                setattr(self, key, data[key])
        self.save_date = datetime.utcnow()



class DagobahLogTask(Base):
    __tablename__ = 'dagobah_log_task'

    id = Column(Integer, primary_key=True)
    log_id = Column(Integer, ForeignKey('dagobah_log.id'), index=True)
    name = Column(String(1000), nullable=False)
    start_time = Column(DateTime)
    complete_time = Column(DateTime)
    success = Column(String(30))
    return_code = Column(Integer)
    stdout = Column(String(STREAM_LOG_SIZE))
    stderr = Column(String(STREAM_LOG_SIZE))
    save_date = Column(DateTime)

    def __init__(self, name):
        self.name = name
        self.save_date = datetime.utcnow()


    def __repr__(self):
        return "<SQLite:DagobahLogTask (%d)>" % self.id


    @property
    def json(self):
        return {'success': self.success,
                'return_code': self.return_code,
                'complete_time': self.complete_time,
                'stdout': self.stdout,
                'stderr': self.stderr}


    def update_from_dict(self, data):
        for key in ['name', 'start_time', 'complete_time',
                    'success', 'return_code', 'stdout',
                    'stderr']:
            if key in data:
                setattr(self, key, data[key])
        self.save_date = datetime.utcnow()



""" Basic HTML email template for sending out Dagobah communications. """

from datetime import datetime
from email.MIMEMultipart import MIMEMultipart
from email.mime.text import MIMEText

import premailer

from dagobah.email.common import EmailTemplate

class BasicEmail(EmailTemplate):

    def send_job_completed(self, data):

        self._format_job_dict(data)
        for task in data.get('tasks', []):
            self._format_task_dict(task)

        template = self._get_template('basic', 'job_completed.html')\
            .render(job=data)
        css = self._get_template('basic', 'job_completed.css').render()

        self.message = self._merge_templates(template, css)
        self._construct_and_send('Job Completed: %s' % data.get('name', None))


    def send_job_failed(self, data):

        self._format_job_dict(data)
        for task in data.get('tasks', []):
            self._format_task_dict(task)

        template = self._get_template('basic', 'job_failed.html')\
            .render(job=data)
        css = self._get_template('basic', 'job_failed.css').render()

        self.message = self._merge_templates(template, css)
        self._construct_and_send('Job Failed: %s' % data.get('name', None))


    def send_task_failed(self, data):

        self._format_task_dict(data)

        template = self._get_template('basic', 'task_failed.html')\
            .render(task=data)
        css = self._get_template('basic', 'task_failed.css').render()

        self.message = self._merge_templates(template, css)
        self._construct_and_send('Task Failed: %s' % data.get('name', None))


    def _format_job_dict(self, job):
        job['next_run'] = self._format_date(job.get('next_run', None))


    def _format_task_dict(self, task):
        success_lu = {None: 'Not executed', True: 'Success',
                      False: 'Failed'}
        task['started_at'] = self._format_date(task.get('started_at', None))
        task['completed_at'] = self._format_date(task.get('completed_at', None))
        task['success'] = success_lu[task.get('success', None)]


    def _merge_templates(self, html, css):
        message = MIMEMultipart()
        content = premailer.transform(''.join(['<html><style>', css, '</style>',
                                               html, '</html>']))
        message.attach(MIMEText(content, 'html'))
        return message


    def _format_date(self, in_date):
        if (not in_date) or (not isinstance(in_date, datetime)):
            return in_date
        return in_date.strftime('%Y-%m-%d %H:%M:%S UTC')



""" Authentication classes and views for Dagobahd. """

from datetime import datetime, timedelta

from flask import render_template, request, url_for, redirect
from flask_login import UserMixin, login_user, logout_user, login_required

from dagobah.daemon.app import app, login_manager

class User(UserMixin):
    def get_id(self):
        return 1

SingleAuthUser = User()


@login_manager.user_loader
def load_user(userid):
    return SingleAuthUser


@app.route('/login', methods=['GET'])
def login():
    return render_template('login.html', alert=request.args.get('alert'))


@app.route('/do-login', methods=['POST'])
def do_login():
    """ Attempt to auth using single login. Rate limited at the site level. """

    dt_filter = lambda x: x >= datetime.utcnow() - timedelta(seconds=60)
    app.config['AUTH_ATTEMPTS'] = filter(dt_filter, app.config['AUTH_ATTEMPTS'])

    if len(app.config['AUTH_ATTEMPTS']) > app.config['AUTH_RATE_LIMIT']:
        return redirect(url_for('login',
                                alert="Rate limit exceeded. Try again in 60 seconds."))

    if request.form.get('password') == app.config['APP_PASSWORD']:
        login_user(SingleAuthUser)
        return redirect('/')

    app.config['AUTH_ATTEMPTS'].append(datetime.utcnow())
    return redirect(url_for('login', alert="Incorrect password."))


@app.route('/do-logout', methods=['GET', 'POST'])
@login_required
def do_logout():
    logout_user()
    return redirect(url_for('login'))



""" Utility functions for the Dagobah daemon. """

from datetime import date, datetime

from flask import request, json, Response, abort
from functools import wraps

try:
    from pymongo.objectid import ObjectId
except ImportError:
    try:
        from bson import ObjectId
    except ImportError:
        pass

from dagobah.core import DagobahError


class DagobahEncoder(json.JSONEncoder):
    def default(self, obj):

        try:
            if isinstance(obj, ObjectId):
                return str(obj)
        except NameError:
            pass

        if isinstance(obj, datetime) or isinstance(obj, date):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


def jsonify(*args, **kwargs):
    return Response(json.dumps(dict(*args, **kwargs),
                               cls=DagobahEncoder,
                               indent=2),
                    mimetype='application/json')


def api_call(fn):
    """ Returns function result in API format if requested from an API
    endpoint """

    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            result = fn(*args, **kwargs)
        except DagobahError:
            if request and request.endpoint == fn.__name__:
                abort(400)
            raise
        except Exception as e:
            print e
            raise

        if request and request.endpoint == fn.__name__:
            status_code = None
            try:
                if result and '_status' in result:
                    status_code = result['_status']
                    del result['_status']
            except TypeError:
                pass

            if isinstance(result, dict):
                if 'result' in result:
                    return jsonify(status=status_code if status_code else 200,
                                   **result)
                else:
                    return jsonify(status=status_code if status_code else 200,
                                   result=result)
            else:
                return jsonify(status=status_code if status_code else 200,
                               result=result)

        else:
            return result

    return wrapper


def validate_dict(in_dict, **kwargs):
    """ Returns Boolean of whether given dict conforms to type specifications
    given in kwargs. """

    if not isinstance(in_dict, dict):
        raise ValueError('requires a dictionary')

    for key, value in kwargs.iteritems():

        if key == 'required':
            for required_key in value:
                if required_key not in in_dict:
                    return False

        elif key not in in_dict:
            continue

        elif value == bool:

            in_dict[key] = (True
                            if str(in_dict[key]).lower() == 'true'
                            else False)

        else:

            if (isinstance(in_dict[key], list) and
                len(in_dict[key]) == 1 and
                value != list):
                in_dict[key] = in_dict[key][0]

            try:
                if key in in_dict:
                    in_dict[key] = value(in_dict[key])
            except ValueError:
                return False

    return True



from setuptools import setup

setup(name='dagobah',
      version='0.1.2',
      description='Simple DAG-based job scheduler',
      url='http://github.com/tthieman/dagobah',
      author='Travis Thieman',
      author_email='travis.thieman@gmail.com',
      license='WTFPL',
      packages=['dagobah',
                'dagobah.backend',
                'dagobah.core',
                'dagobah.daemon',
                'dagobah.email'],
      package_data={'dagobah': ['email/templates/basic/*',
                                'daemon/static/css/*',
                                'daemon/static/js/*',
                                'daemon/static/img/*',
                                'daemon/static/lib/*.js',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/*.js',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/*.ks',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/affix/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/alert/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/button/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/carousel/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/collapse/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/dropdown/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/modal/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/popover/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/scrollspy/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/tab/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/tooltip/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/transition/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/bootstrap/typeahead/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/extras/fontawesome/font/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/retina/*',
                                'daemon/static/lib/Kickstrap1.3.2/Kickstrap/apps/tinygrowl/*',
                                'daemon/templates/*',
                                'daemon/dagobahd.yml']},
      install_requires=['croniter', 'pyyaml', 'flask', 'premailer', 'flask-login'],
      test_suite='nose.collector',
      tests_require=['nose', 'pymongo'],
      entry_points={'console_scripts':
                    ['dagobahd = dagobah.daemon.app:daemon_entrypoint',
                     'echo_dagobah_conf = dagobah.daemon.daemon:print_standard_conf']
                    },
      zip_safe=False)



""" Base Backend class inherited by specific implementations. """

import os
import binascii

class BaseBackend(object):
    """ Base class for prototypes and compound functions.

    This is also used as the default Backend if the user
    does not specify anything at runtime. In this case,
    calls will proceed normally, but the methods here
    will not persist anything permanently.
    """

    def __init__(self):
        pass


    def __repr__(self):
        return '<BaseBackend>'


    def get_known_dagobah_ids(self):
        return []


    def get_new_dagobah_id(self):
        return binascii.hexlify(os.urandom(16))


    def get_new_job_id(self):
        return binascii.hexlify(os.urandom(16))


    def get_new_log_id(self):
        return binascii.hexlify(os.urandom(16))


    def get_dagobah_json(self, dagobah_id):
        return


    def commit_dagobah(self, dagobah_json):
        return


    def delete_dagobah(self, dagobah_id):
        return


    def commit_job(self, job_json):
        pass


    def delete_job(self, job_name):
        pass


    def commit_log(self, log_json):
        pass


    def get_latest_run_log(self, job_id, task_name):
        return {}


    def acquire_lock(self):
        return


    def release_lock(self):
        return
